#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

import os
import sys
import shutil
import enum
import copy
import json
import gzip
import numpy as np

import pyomo.environ as pyo
from pyomo.common.collections import ComponentSet, ComponentMap
from pyomo.core.expr.visitor import identify_variables
import pyomo.dae as pyodae
from pyomo.common import Executable
from pyomo.dae.flatten import flatten_dae_components, slice_component_along_sets
from pyomo.util.subsystems import (
    TemporarySubsystemManager,
    create_subsystem_block,
)
from pyomo.solvers.plugins.solvers.ASL import ASL
from pyomo.common.tempfiles import TempfileManager
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from pyomo.common.deprecation import deprecation_warning

import idaes
import idaes.logger as idaeslog
import idaes.config as icfg


# Importing a few things here so that they are cached
# pylint: disable=unused-import
# pylint: disable=import-outside-toplevel
# pylint: disable=protected-access


def petsc_binary_io():
    if petsc_binary_io.PetscBinaryIOTrajectory is not None:
        return petsc_binary_io.PetscBinaryIOTrajectory

    # First see if the python IO helpers are directly importable
    try:
        import PetscBinaryIOTrajectory
        import PetscBinaryIO

        petsc_binary_io.PetscBinaryIOTrajectory = PetscBinaryIOTrajectory
        return PetscBinaryIOTrajectory
    except ImportError:
        pass
    # Next, look for a 'petscpy' directory alongside the 'petsc'
    # executable: first look at the petsc we found on the path, then
    # look in the IDAES bin dir.  Casting the Executable path to a
    # string will map None to '' in the case where there is no petsc
    # executable found.
    for petsc_exe_dir in (
        os.path.dirname(str(Executable("petsc").path())),
        icfg.bin_directory,
    ):
        if not petsc_exe_dir:
            continue
        petscpy_dir = os.path.join(petsc_exe_dir, "petscpy")
        try:
            sys.path.insert(0, petscpy_dir)
            import PetscBinaryIOTrajectory
            import PetscBinaryIO

            # Import the petsc_conf so that it is cached in sys.modules
            # (because we will be removing petscpy from sys.path)
            import petsc_conf

            petsc_binary_io.PetscBinaryIOTrajectory = PetscBinaryIOTrajectory
            return PetscBinaryIOTrajectory
        except ImportError:
            pass
        finally:
            sys.path.remove(petscpy_dir)
    return None


petsc_binary_io.PetscBinaryIOTrajectory = None


class DaeVarTypes(enum.IntEnum):
    """Enum DAE variable type for suffix"""

    ALGEBRAIC = 0
    DIFFERENTIAL = 1
    DERIVATIVE = 2
    TIME = 3


@pyo.SolverFactory.register("petsc", doc="ASL PETSc interface")
class Petsc(ASL):
    """ASL solver plugin for the PETSc solver.  This adds the option to use an
    alternative executable batch file to run the solver through the WSL."""

    def __init__(self, **kwds):
        super().__init__(**kwds)
        self.options.solver = "petsc"

    def _default_executable(self):
        """In addition to looking for the petsc executable, optionally check for
        a WSL batch file on Windows. Users could potentially also compile a
        cygwin executable on Windows, so WSL isn't the only option, but it is the
        easiest for Windows."""
        executable = Executable("petsc")
        if not executable:
            raise RuntimeError("No PETSc executable found.")
        return executable.path()


@pyo.SolverFactory.register("petsc_snes", doc="ASL PETSc SNES interface")
class PetscSNES(Petsc):
    """PETSc solver plugin that sets options for SNES solver.  This turns on
    SNES monitoring, and checks the config for default options"""

    def __init__(self, **kwds):
        if "options" in kwds and kwds["options"] is not None:
            kwds["options"] = copy.deepcopy(kwds["options"])
        else:
            kwds["options"] = {}
        kwds["options"]["--snes_monitor"] = ""
        if "petsc_snes" in idaes.cfg:
            if "options" in idaes.cfg["petsc_snes"]:
                default_options = dict(idaes.cfg["petsc_snes"]["options"])
                default_options.update(kwds["options"])
                kwds["options"] = default_options
        super().__init__(**kwds)
        self.options.solver = "petsc_snes"


@pyo.SolverFactory.register("petsc_ts", doc="ASL PETSc TS interface")
class PetscTS(Petsc):
    """PETSc solver plugin that sets options for TS solver.  This turns on
    TS monitoring, sets the DAE flag, and checks the config for default options.
    """

    def __init__(self, **kwds):
        self._ts_vars_stub = kwds.pop("vars_stub", "tmp_vars_stub")
        if "options" in kwds and kwds["options"] is not None:
            kwds["options"] = copy.deepcopy(kwds["options"])
        else:
            kwds["options"] = {}
        # Force some options.
        kwds["options"]["--dae_solve"] = ""  # is DAE solver
        kwds["options"]["--ts_monitor"] = ""  # show TS solver progress
        # We're assuming trajectory will be written in the visualization
        # style, so just set that here.
        kwds["options"]["--ts_trajectory_type"] = "visualization"
        # Get the options set in the IDAES config
        if "petsc_ts" in idaes.cfg:
            if "options" in idaes.cfg["petsc_ts"]:
                default_options = dict(idaes.cfg["petsc_ts"]["options"])
                default_options.update(kwds["options"])
                kwds["options"] = default_options
        super().__init__(**kwds)
        self.options.solver = "petsc_ts"

    def _postsolve(self):
        stub = os.path.splitext(self._soln_file)[0]
        # There is a type file created by the solver to give the variable types
        # this is needed to read the trajectory data, and we want to include it
        # with other tmp files
        typ_file = stub + ".typ"
        TempfileManager.add_tempfile(typ_file)
        # If trajectory is saved, copy the col and typ files to
        # the working directory. These files are needed to get the names and
        # types of variables and to make sense of trajectory data.
        if self.options.get("--ts_save_trajectory", 0):
            try:
                shutil.copyfile(f"{stub}.col", f"{self._ts_vars_stub}.col")
            except Exception:  # pylint: disable=W0703
                pass
            try:
                shutil.copyfile(f"{stub}.typ", f"{self._ts_vars_stub}.typ")
            except Exception:  # pylint: disable=W0703
                pass
        return ASL._postsolve(self)


@pyo.SolverFactory.register("petsc_tao", doc="ASL PETSc TAO interface")
class PetscTAO(Petsc):
    """This is a place holder for optimization solvers"""

    # Placeholder class, skip pylint checks
    # pylint: disable=W0231
    def __init__(self, **kwds):
        raise NotImplementedError(
            "The PETSc TAO interface has not yet been implemented"
        )


def petsc_available():
    """Check if the IDAES AMPL solver wrapper for PETSc is available.

    Args:
        None

    Returns (bool):
        True if PETSc is available
    """
    solver = pyo.SolverFactory("petsc")
    if solver is not None:
        try:
            return solver.available()
        except RuntimeError:
            return False
    return False


def _copy_time(time_vars, t_from, t_to):
    """PRIVATE FUNCTION:

    This is used on the flattened (indexed only by time) variable
    representations to copy variable values that are unfixed at the "to" time
    from the value at the "from" time. The PETSc DAE solver uses the initial
    variable values as the initial condition, so this is used to copy the
    previous time in as the initial condition for the next step.

    Args:
        time_vars (list): list of variables or references to variables indexed
            only by time
        t_from (float): time point to copy from
        t_to (float): time point to copy to, only unfixed vars will be
            overwritten

    Returns:
        None
    """
    for v in time_vars:
        if not v[t_to].fixed:
            v[t_to].value = v[t_from].value


def find_discretization_equations(m, time):
    """This returns a list of time discretization equations. Since we aren't
    solving the whole time period simultaneously, we'll want to deactivate
    these constraints.

    Args:
        m (Block): model or block to search for constraints
        time (ContinuousSet):

    Returns:
        list of time discretization constraints
    """
    disc_eqns = []
    for var in m.component_objects(pyo.Var):
        if isinstance(var, pyodae.DerivativeVar):
            cont_set_set = ComponentSet(var.get_continuousset_list())
            if time in ComponentSet(cont_set_set):
                if len(cont_set_set) > 1:
                    raise NotImplementedError(
                        f"IDAES presently does not support PETSc for second order or higher derivatives like {var.name} "
                        "that are differentiated at least once with respect to time. Please reformulate your model so "
                        "it does not contain such a derivative (such as by introducing intermediate variables)."
                    )
                parent = var.parent_block()
                name = var.local_name + "_disc_eq"
                disc_eq = getattr(parent, name)
                disc_eqns.append(disc_eq)
    return disc_eqns


def _set_dae_suffixes_from_variables(m, variables, deriv_diff_map):
    """Write suffixes used by the solver to identify different variable types
    and associated derivative and differential variables.

    Args:
        m: model to search for variables and write suffixes to
        variables (list): List of time indexed variables at a specific time
            point
        deriv_diff_map (ComponentMap): Maps DerivativeVar data objects to
            differential variable data objects

    Returns:
        None
    """
    # The dae_suffix provides the solver information about variables types
    # algebraic, differential, derivative, or time, see DaeVarTypes
    m.dae_suffix = pyo.Suffix(
        direction=pyo.Suffix.EXPORT,
        datatype=pyo.Suffix.INT,
    )
    # The dae_link suffix provides the solver a link between the differential
    # and derivative variable, by assigning the pair a unique integer index.
    m.dae_link = pyo.Suffix(
        direction=pyo.Suffix.EXPORT,
        datatype=pyo.Suffix.INT,
    )
    dae_var_link_index = 1
    differential_vars = []
    i = 0
    for var in variables:
        if var in deriv_diff_map:
            deriv = var
            diffvar = deriv_diff_map[deriv]
            m.dae_suffix[diffvar] = int(DaeVarTypes.DIFFERENTIAL)
            m.dae_suffix[deriv] = int(DaeVarTypes.DERIVATIVE)
            m.dae_link[diffvar] = dae_var_link_index
            m.dae_link[deriv] = dae_var_link_index
            i += 1
            dae_var_link_index += 1
            if not diffvar.fixed:
                differential_vars.append(diffvar)
            else:
                raise RuntimeError(
                    f"Problem cannot contain a fixed differential variable and "
                    f"unfixed derivative. Consider either fixing the "
                    f"corresponding derivative or adding a constraint for the "
                    f"differential variable {diffvar} possibly using an "
                    f"explicit time variable."
                )
    return differential_vars


def _get_derivative_differential_data_map(m, time):
    """Get a map from data objects of derivative variables to the
    corresponding data objects of differential variables.

    Args:
        m: Model to search for DerivativeVars
        time: Set with respect to which DerivativeVars must be differentiated

    Returns:
        (ComponentMap): Map from derivative data objects to differential
            data objects
    """
    # Get corresponding derivative and differential data objects,
    # with no attention paid to fixed or active status.
    deriv_diff_list = []
    for var in m.component_objects(pyo.Var):
        if isinstance(var, pyodae.DerivativeVar) and time in ComponentSet(
            var.get_continuousset_list()
        ):
            deriv = var
            diffvar = deriv.get_state_var()
            for idx in var:
                if deriv[idx].fixed and pyo.value(abs(deriv[idx])) > 1e-10:
                    raise RuntimeError(
                        f"{deriv[idx]} is fixed to a nonzero value "
                        f"{pyo.value(deriv[idx])}. This is "
                        f"most likely a modeling error. Instead of fixing the "
                        f"derivative consider adding a constraint like "
                        f"dxdt = constant"
                    )
                deriv_diff_list.append((deriv[idx], diffvar[idx]))

    # Get unfixed variables in active constraints
    active_con_vars = ComponentSet()
    for con in m.component_data_objects(pyo.Constraint, active=True):
        for var in identify_variables(con.expr, include_fixed=False):
            active_con_vars.add(var)

    # Filter out derivatives that are fixed or not in an active constraint
    filtered_deriv_diff_list = []
    for deriv, diff in deriv_diff_list:
        if deriv in active_con_vars:
            filtered_deriv_diff_list.append((deriv, diff))
    return ComponentMap(filtered_deriv_diff_list)


def _sub_problem_scaling_suffix(m, t_block):
    """Copy scaling factors from the full model to the submodel.  This assumes
    the scaling suffixes could be in two places.  First check the parent block
    of the component (typical place for idaes models) then check the top-level
    model.  The top level model will take precedence.
    """
    if not hasattr(t_block, "scaling_factor"):
        # if the subsystem block doesn't already have a scaling suffix, make one
        t_block.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)
    # first check the parent block for scaling factors
    for c in t_block.component_data_objects((pyo.Var, pyo.Constraint)):
        if hasattr(c.parent_block(), "scaling_factor"):
            if c in c.parent_block().scaling_factor:
                t_block.scaling_factor[c] = c.parent_block().scaling_factor[c]
    # now check the top level model
    if hasattr(m, "scaling_factor"):
        for c in t_block.component_data_objects((pyo.Var, pyo.Constraint)):
            if c in m.scaling_factor:
                t_block.scaling_factor[c] = m.scaling_factor[c]


class PetscDAEResults(object):
    """This class stores the results of ``petsc_dae_by_time_element()`` it has
    two attributes ``results`` and ``trajectory``.  Results is a list of Pyomo
    solver results objects.  Since this PETSc DAE solver utility solves the
    initial conditions first then may integrate over the time domain in one or
    more steps, the results of multiple solves are returned.  The trajectory is
    a ``PetscTrajectory`` object which gives the full trajectory of the solution
    for all time steps taken by the PETSc solver. This is generally finer than
    the Pyomo.DAE discretization.
    """

    def __init__(self, results=None, trajectory=None):
        self.results = results
        self.trajectory = trajectory


def petsc_dae_by_time_element(
    m,
    time,
    timevar=None,
    initial_constraints=None,
    initial_variables=None,
    detect_initial=True,
    skip_initial=False,
    initial_solver="petsc_snes",
    initial_solver_options=None,
    ts_options=None,
    keepfiles=False,
    symbolic_solver_labels=True,
    between=None,
    interpolate=True,
    calculate_derivatives=False,
    previous_trajectory=None,
    representative_time=None,
    snes_options=None,
):
    """Solve a DAE problem step by step using the PETSc DAE solver.  This
    integrates from one time point to the next.

    Args:
        m (Block): Pyomo model to solve
        time (ContinuousSet): Time set
        timevar (Var): Optional specification of a time variable, which can be
            used to write constraints that are an explicit function of time.
        initial_constraints (list): Constraints to solve in the initial
            condition solve step.  Since the time-indexed constraints are picked
            up automatically, this generally includes non-time-indexed
            constraints.
        initial_variables (list): This is a list of variables to fix after the
            initial condition solve step.  If these variables were originally
            unfixed, they will be unfixed at the end of the solve. This usually
            includes non-time-indexed variables that are calculated along with
            the initial conditions.
        detect_initial (bool): If True, add non-time-indexed variables and
            constraints to initial_variables and initial_constraints.
        skip_initial (bool): Don't do the initial condition calculation step,
            and assume that the initial condition values have already been
            calculated. This can be useful, for example, if you read initial
            conditions from a separately solved steady state problem, or
            otherwise know the initial conditions.
        initial_solver (str): default=petsc_snes, the nonlinear equations solver
            to use for the initial conditions (e.g. petsc_snes, ipopt, ...).
        initial_solver_options (dict): nonlinear equation solver options
        ts_options (dict): PETSc time-stepping solver options
        keepfiles (bool): pass to keepfiles arg for solvers
        symbolic_solver_labels (bool): pass to symbolic_solver_labels argument
            for solvers. If you want to read trajectory data from the
            time-stepping solver, this should be True.
        between (list or tuple): List of time points to integrate between. If
            None use all time points in the model.
        interpolate (bool): if True and trajectory is read, interpolate model
            values from the trajectory
        calculate_derivatives: (bool) if True, calculate the derivative values
            based on the values of the differential variables in the discretized
            Pyomo model.
        previous_trajectory: (PetscTrajectory) Trajectory from previous integration
            of this model. New results will be appended to this trajectory object.
        representative_time: (element of Between) Time when all equations necessary to solve DAE are active. Often equations need
            to be deactivated for the initial condition problem (for example, mole fractions summing to one)
            because state variables (the individual mole fractions) are fixed. representative_time is a time
            after the initial condition problem is solved and these equations are reactivated. Note that the
            equations not active at this point are excluded from the DAE at future points. If no
            representative_time is specified, it is assumed to be the second element of between.
            Must be an element of between.
        snes_options (dict): [DEPRECATED in favor of initial_solver_options] nonlinear equation solver options

    Returns (PetscDAEResults):
        See PetscDAEResults documentation for more information.
    """
    if snes_options is not None and initial_solver_options is not None:
        raise RuntimeError(
            "Both (deprecated) snes_options and initial_solver_options keyword arguments were specified. "
            "Please specify your initial solver options using only the initial_solver_options argument."
        )
    if snes_options is not None:
        _log = idaeslog.getLogger(__name__)
        deprecation_warning(
            msg="Keyword argument snes_options has been DEPRECATED in favor of initial_solver_options.",
            logger=_log,
            version="2.2.0",
            remove_in="3.0.0",
        )
        initial_solver_options = snes_options

    if interpolate:
        if ts_options is None:
            ts_options = {}
        ts_options["--ts_save_trajectory"] = 1
    if timevar:
        for t in time:
            timevar[t].set_value(t)
    if between is None:
        between = time
    else:
        bad_times = between - time
        if bad_times:
            raise RuntimeError(
                "Elements of the 'between' argument must be in the time set"
            )
        between = pyo.Set(initialize=sorted(between))
        between.construct()

    # Make sure that time is a ContinuousSet that has been discretized
    if not isinstance(time, pyodae.ContinuousSet):
        raise RuntimeError("Argument time is not a Pyomo ContinuousSet")

    if "scheme" not in time.get_discretization_info():
        raise RuntimeError("The ContinuousSet time has not been discretized")

    solve_log = idaeslog.getSolveLogger("petsc-dae")

    # Need a representative time for flatten_dae_components to work, because otherwise things like
    # sums of mole fraction constraints being deactivated at t0 (because all the mole fractions are fixed),
    # etc., cause flatten_dae_components to miss a constraint that is active at t>t0
    if representative_time is None:
        representative_time = between.at(2)  # Two because Pyomo sets start at one
    elif representative_time not in between:
        raise RuntimeError("representative_time is not element of between.")

    regular_vars, time_vars = flatten_dae_components(
        m, time, pyo.Var, active=True, indices=(representative_time,)
    )
    regular_cons, time_cons = flatten_dae_components(
        m, time, pyo.Constraint, active=True, indices=(representative_time,)
    )
    tdisc = find_discretization_equations(m, time)

    solver_dae = pyo.SolverFactory("petsc_ts", options=ts_options)
    save_trajectory = solver_dae.options.get("--ts_save_trajectory", 0)

    # First calculate the initial conditions and non-time-indexed constraints
    res_list = []
    t0 = between.first()
    # list of variables to add to initial condition problem
    if initial_variables is None:
        initial_variables = []
    if detect_initial:
        rvset = ComponentSet(regular_vars)
        ivset = ComponentSet(initial_variables)
        initial_variables = list(ivset | rvset)

    if not skip_initial:
        # Nonlinear equation solver for initial conditions
        initial_solver_obj = pyo.SolverFactory(
            initial_solver, options=initial_solver_options
        )
        # list of constraints to add to the initial condition problem
        if initial_constraints is None:
            initial_constraints = []

        if detect_initial:
            # If detect_initial, solve the non-time-indexed variables and
            # constraints with the initial conditions
            rcset = ComponentSet(regular_cons)
            icset = ComponentSet(initial_constraints)
            initial_constraints = list(icset | rcset)

        with TemporarySubsystemManager(to_deactivate=tdisc):
            constraints = [
                con[t0] for con in time_cons if t0 in con
            ] + initial_constraints
            variables = [var[t0] for var in time_vars] + initial_variables
            if len(constraints) > 0:
                # if the initial condition is specified and there are no
                # initial constraints, don't try to solve.
                t_block = create_subsystem_block(
                    constraints,
                    variables,
                )
                # set up the scaling factor suffix
                _sub_problem_scaling_suffix(m, t_block)
                with idaeslog.solver_log(solve_log, idaeslog.INFO) as slc:
                    res = initial_solver_obj.solve(
                        t_block,
                        tee=slc.tee,
                        symbolic_solver_labels=symbolic_solver_labels,
                    )
                res_list.append(res)

    tprev = t0
    tj = previous_trajectory
    if tj is not None:
        variables_prev = [var[t0] for var in time_vars]

    with TemporarySubsystemManager(
        to_deactivate=tdisc,
        to_fix=initial_variables,
    ):
        # Solver time steps
        deriv_diff_map = _get_derivative_differential_data_map(m, time)
        for t in between:
            if t == between.first():
                # t == between.first() was handled above
                continue
            constraints = [con[t] for con in time_cons if t in con]
            variables = [var[t] for var in time_vars]
            # Create a temporary block with references to original constraints
            # and variables so we can integrate this "subsystem" without
            # altering the rest of the model.
            t_block = create_subsystem_block(constraints, variables)
            differential_vars = _set_dae_suffixes_from_variables(
                t_block,
                variables,
                deriv_diff_map,
            )
            # We need to check if there are derivatives in the problem before
            # sending this to the solver.  We'll assume that if you are using
            # this and don't have any differential equations, you are making a
            # mistake.
            if len(differential_vars) < 1:
                raise RuntimeError(
                    f"No differential equations found at t = {t}, not a DAE"
                )
            if timevar is not None:
                t_block.dae_suffix[timevar[t]] = int(DaeVarTypes.TIME)
            # Set up the scaling factor suffix
            _sub_problem_scaling_suffix(m, t_block)
            # Take initial conditions for this step from the result of previous
            _copy_time(time_vars, tprev, t)
            with idaeslog.solver_log(solve_log, idaeslog.INFO) as slc:
                res = solver_dae.solve(
                    t_block,
                    tee=slc.tee,
                    keepfiles=keepfiles,
                    symbolic_solver_labels=symbolic_solver_labels,
                    export_nonlinear_variables=differential_vars,
                    options={"--ts_init_time": tprev, "--ts_max_time": t},
                )
            if save_trajectory:
                tj_prev = tj
                tj = PetscTrajectory(
                    stub="tmp_vars_stub",
                    delete_on_read=True,
                    unscale=True,
                    model=t_block,
                )
                # add fixed vars to the trajectory. this does two things 1)
                # helps users looking for fixed var trajectory and 2) lets
                # us concatenate trajectories with section that are mixed fixed
                # and unfixed
                for i, v in enumerate(variables):
                    if isinstance(v.parent_component(), pyodae.DerivativeVar):
                        continue  # skip derivative vars
                    try:
                        vec = tj.get_vec(v)
                    except KeyError:
                        tj._set_vec(v, [pyo.value(v)] * len(tj.time))
                if tj_prev is not None:
                    # due to the way variables is generated we know variables
                    # have corresponding positions in the list
                    no_repeat = set()
                    for i, v in enumerate(variables):
                        vp = variables_prev[i]
                        if id(v) in no_repeat:
                            continue  # variables can be repeated in list
                        if isinstance(v.parent_component(), pyodae.DerivativeVar):
                            continue  # skip derivative vars
                        no_repeat.add(id(v))
                        # We'll add fixed vars in case they aren't fixed in
                        # another section. Fixed vars don't go to the solver
                        # so they don't show up in the trajectory data
                        vec = tj.get_vec(v)
                        vec_prev = tj_prev.get_vec(vp)
                        tj._set_vec(v, vec_prev + vec)
                    tj._set_time_vec(tj_prev.time + tj.time)
                variables_prev = variables
            tprev = t
            res_list.append(res)
        # If the interpolation option is True and the trajectory is available
        # interpolate the values any skipped time points from the trajectory
        if interpolate and tj is not None:
            t0 = between.first()
            tlast = between.last()
            itime = [t for t in time if not (t < t0 or t > tlast)]
            no_repeat = set()
            if timevar is not None:
                for t in itime:
                    timevar[t] = t
                no_repeat.add(id(timevar[tlast]))
            for var in time_vars:
                if id(var[tlast]) in no_repeat:
                    continue
                no_repeat.add(id(var[tlast]))
                if isinstance(var[t0].parent_component(), pyodae.DerivativeVar):
                    continue  # skip derivative vars
                vec = tj.interpolate_vec(itime, var[tlast])
                for i, t in enumerate(itime):
                    if t in between:
                        # Time is already set
                        continue
                    if not var[t].fixed:
                        # May not have trajectory from fixed variables and they
                        # shouldn't change anyway, so only set not fixed vars
                        var[t].value = vec[i]
    if calculate_derivatives:
        # the petsc solver interface does not currently return time
        # derivatives, and if it did, they would be estimated based on a
        # smaller time step.  This option uses Pyomo.DAE's discretization
        # equations to calculate the time derivative values
        calculate_time_derivatives(m, time, between=between)
    # return the solver results and trajectory if available
    return PetscDAEResults(results=res_list, trajectory=tj)


def calculate_time_derivatives(m, time, between=None):
    """Calculate the derivative values from the discretization equations.

    Args:
        m (Block): Pyomo model block
        time (ContinuousSet): Time set

    Returns:
        None
    """
    # Leave between an optional argument for backwards compatibility
    if between is None:
        between = time
    for var in m.component_objects(pyo.Var):
        if isinstance(var, pyodae.DerivativeVar):
            if time in ComponentSet(var.get_continuousset_list()):
                parent_block = var.parent_block()
                disc_eq = getattr(parent_block, var.local_name + "_disc_eq")

                deriv_dict = dict(
                    (key, pyo.Reference(slc))
                    for key, slc in slice_component_along_sets(var, (time,))
                )
                disc_dict = dict(
                    (key, pyo.Reference(slc))
                    for key, slc in slice_component_along_sets(disc_eq, (time,))
                )

                for key, deriv in deriv_dict.items():
                    # state = state_dict[key]
                    disc_eq = disc_dict[key]
                    for t in time:
                        if t < between.first() or t > between.last():
                            # Outside of integration range, skip calculation
                            continue
                        old_value = deriv[t].value
                        try:
                            # TODO This calculates the value of the derivative even
                            # if one of the state var values is from outside the
                            # integration range, so long as it's initialized. Is
                            # this the desired behavior?
                            if disc_eq[t].active and not deriv[t].fixed:
                                deriv[t].value = 0  # Make sure there is a value
                                calculate_variable_from_constraint(deriv[t], disc_eq[t])
                        except KeyError as err:
                            # Discretization and continuity equations may or may not exist at the first or last time
                            # points depending on the method. Backwards skips first, forwards skips last, central skips
                            # both (which means the user needs to provide additional equations)
                            if t == time.first() or t == time.last():
                                pass
                            else:
                                raise err
                        except ValueError as err:
                            # At edges of between, it's unclear which adjacent
                            # values of state variables have been populated.
                            # Therefore we might get hit with value errors.
                            if t == between.first() or t == between.last():
                                # Reset deriv value to old value
                                if disc_eq[t].active and not deriv[t].fixed:
                                    deriv[t].value = old_value
                            else:
                                raise err


class PetscTrajectory(object):
    def __init__(
        self,
        stub=None,
        vecs=None,
        json=None,
        pth=None,
        vis_dir="Visualization-data",
        delete_on_read=False,
        unscale=None,
        model=None,
        no_read=False,
    ):
        """Class to read PETSc TS solver trajectory data.  This can either read
        PETSc output by providing the ``stub`` argument, a trajectory dict by
        providing ``vecs`` or a json file by providing ``json``.

        Args:
            stub (str): file name stub for variable info
            pth (str): path where variable info and trajectory data are stored
                if None, use current working directory
            vis_dir (str): subdirectory where visualization data is stored
            delete_on_read (bool): if true delete trajectory data after reading
            unscale (bool or Block): if True and model is specified, use model
                to obtain scale factors and unscale the trajectory. If a Block
                is specified use the specified block to unscale the model. If
                False or None do not unscale.
            model (Block): if specified use for unscaling
            no_read (bool): if True make an uninitialized trajectory object
        """
        if no_read:
            return
        if petsc_binary_io() is None and stub is not None:
            raise RuntimeError("PetscBinaryIOTrajectory could not be imported")
        # if unscale is True, use model as scale factor source
        if model is not None and unscale is True:
            unscale = model
        self.model = model
        self.id_map = {}
        if pth is not None:
            stub = os.path.join(pth, stub)
            vis_dir = os.path.join(pth, vis_dir)
        if stub is not None:
            self.stub = stub
            self.vis_dir = vis_dir
            self.path = pth
            self.unscale = unscale
            self._read()
            if delete_on_read:
                self.delete_files()
            if unscale is not None:
                self._unscale(unscale)
        elif vecs is not None:
            self.vecs = vecs
            self.time = vecs["_time"]
        elif json is not None:
            self.from_json(json)
        else:
            raise RuntimeError("To read trajectory, provide stub, vecs, or json")

    def _read(self):
        with open(f"{self.stub}.col") as f:
            names = list(map(str.strip, f.readlines()))
        with open(f"{self.stub}.typ") as f:
            typ = list(map(int, f.readlines()))
        _vars = [name for i, name in enumerate(names) if typ[i] in [0, 1]]
        (t, v, names) = petsc_binary_io().ReadTrajectory("Visualization-data")
        self.time = t
        self.vecs_by_time = v
        self.vecs = dict.fromkeys(_vars, None)
        for k in self.vecs.keys():
            self.vecs[k] = [0] * len(self.time)
        self.vecs["_time"] = list(self.time)
        for i, v in enumerate(_vars):
            for j, vt in enumerate(self.vecs_by_time):
                self.vecs[v][j] = vt[i]

    def _set_vec(self, var, vec):
        try:
            var = self.id_map[id(var)]
        except KeyError:
            var_str = str(var)
            self.id_map[id(var)] = var_str
            var = var_str
        self.vecs[var] = vec

    def _set_time_vec(self, vec):
        self.vecs["_time"] = vec
        self.time = self.vecs["_time"]

    def get_vec(self, var):
        """Return the vector of variable values at each time point for var.

        Args:
            var (str or Var): Variable to get vector for.
            time (Set): Time index set

        Returns (list):
            vector of variable values at each time point

        """
        try:
            var = self.id_map[id(var)]
        except KeyError:
            var_str = str(var)
            self.id_map[id(var)] = var_str
            var = var_str
        return self.vecs[var]

    def get_dt(self):
        """Get a list of time steps

        Args:
            None

        Returns:
            (list)
        """
        dt = [None] * (len(self.time) - 1)
        for i in range(len(self.time) - 1):
            dt[i] = self.time[i + 1] - self.time[i]
        return dt

    def interpolate(self, times):
        """Create a new vector dictionary interpolated at times. This method
        will also extrapolate values outside the original time range, so care
        should be taken not to specify times too far outside the range.

        Args:
            times (list): list of times to interpolate. These must be in
                increasing order.

        Returns (dict):
            Dictionary of vectors for values at interpolated points
        """
        tj = PetscTrajectory(no_read=True)
        tj.id_map = copy.copy(self.id_map)
        tj.vecs = dict.fromkeys(self.vecs.keys(), None)
        tj.vecs["_time"] = copy.copy(times)
        tj.time = tj.vecs["_time"]
        for var in self.vecs:
            if var == "_time":
                continue
            tj.vecs[var] = np.interp(tj.time, self.time, self.vecs[var])
        return tj

    def interpolate_vec(self, times, var):
        """Create a new vector dictionary interpolated at times. This method
        will also extrapolate values outside the original time range, so care
        should be taken not to specify times too far outside the range.

        Args:
            times (list): list of times to interpolate. These must be in
                increasing order.

        Returns (dict):
            Dictionary of vectors for values at interpolated points
        """
        return np.interp(times, self.time, self.get_vec(var))

    def _unscale(self, m):
        """If variable scale factors are used, the solver will see scaled
        variables, and the scaled trajectory will be written. This function
        uses variable scaling factors from the given model to unscale the
        trajectory.

        Args:
            m (Block): model or block to read scale factors from.

        Returns:
            None
        """
        # Variables might show up more than once because of References
        already_scaled = set()
        for var in m.component_data_objects():
            try:
                vec = self.get_vec(var)
            except KeyError:
                continue
            vid = id(var)
            if vid in already_scaled:
                continue
            s = None
            if hasattr(var.parent_block(), "scaling_factor"):
                s = var.parent_block().scaling_factor.get(var, s)
            if hasattr(m, "scaling_factor"):
                s = m.scaling_factor.get(var, s)
            if s is not None:
                for i, x in enumerate(vec):
                    vec[i] = x / s
            already_scaled.add(vid)

    def delete_files(self):
        """Delete the trajectory data and variable information files.

        Args:
            None

        Returns:
            None
        """
        shutil.rmtree(self.vis_dir)
        os.remove(f"{self.stub}.col")
        os.remove(f"{self.stub}.typ")

    def to_json(self, pth):
        """Dump the trajectory data to a json file in the form of a dictionary
        with variable name keys and '_time' with vectors of values at each time.

        Args:
            pth (str): path for json file to write

        Returns:
            None
        """
        if pth.endswith(".gz"):
            with gzip.open(pth, "w") as fp:
                fp.write(json.dumps(self.vecs).encode("utf-8"))
        else:
            with open(pth, "w") as fp:
                json.dump(self.vecs, fp)

    def from_json(self, pth):
        """Read the trajectory data from a json file in the form of a dictionary.

        Args:
            pth (str): path for json file to write

        Returns:
            None
        """
        if pth.endswith(".gz"):
            with gzip.open(pth, "r") as fp:
                self.vecs = json.loads(fp.read())
        else:
            with open(pth, "r") as fp:
                self.vecs = json.load(fp)
        self.time = self.vecs["_time"]



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
CO2-bmimPF6 phase equilibrium package using ideal liquid and vapor.

Example property package using the Generic Property Package Framework.
This exmample shows how to set up a property package to do CO2-Ionic Liquid
phase equilibrium in the generic framework using Peng-robinsons EOS
assumptions along with methods drawn from the pre-built IDAES property
libraries.
"""
# Import Python libraries
import logging

# Import Pyomo units
from pyomo.environ import units as pyunits

# Import IDAES cores
from idaes.core import LiquidPhase, VaporPhase, Component, PhaseType as PT
from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.eos.ceos import Cubic, CubicType
from idaes.models.properties.modular_properties.phase_equil import SmoothVLE
from idaes.models.properties.modular_properties.phase_equil.bubble_dew import (
    LogBubbleDew,
)
from idaes.models.properties.modular_properties.phase_equil.forms import log_fugacity

from idaes.models.properties.modular_properties.pure import RPP4
from idaes.models.properties.modular_properties.pure import NIST

# Set up logger
_log = logging.getLogger(__name__)


# ---------------------------------------------------------------------
# Configuration dictionary for an ideal Benzene-Toluene system

# Data Sources:
# [1] The Properties of Gases and Liquids (1987)
#     4th edition, Chemical Engineering Series - Robert C. Reid
# [2] Engineering Toolbox, https://www.engineeringtoolbox.com
#     Retrieved 1st December, 2019
# [3] Separation of CO2 and H2S using room-temperature ionic liquid [bmim][PF6]
#     Mark B.Shiflett, A.Yokozeki, 2010
# [4] Thermodynamic Properties of Imidazolium-Based Ionic Liquids:  Densities, Heat Capacities, and Enthalpies of Fusion of [bmim][PF6] and [bmim][NTf2]
#     Jacobo Troncoso, Claudio A. Cerdeiriña, Yolanda A. Sanmamed, Luís Romaní, and Luís Paulo N. Rebelo, 2006

configuration = {
    # Specifying components
    "components": {
        "bmimPF6": {
            "type": Component,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "valid_phase_types": PT.liquidPhase,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (284.18e-3, pyunits.kg / pyunits.mol),  # [3]
                "pressure_crit": (24e5, pyunits.Pa),  # [3]
                "temperature_crit": (860, pyunits.K),  # [3]
                "omega": 0.7917,  # [4]
                "cp_mol_ig_comp_coeff": {
                    "A": (3259.5745, pyunits.J / pyunits.mol / pyunits.K),  # [4]
                    "B": (-28.5610, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (0.09354, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (-0.0001000673, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "enth_mol_form_vap_comp_ref": (0.145, pyunits.J / pyunits.mol),  # [4]
                "entr_mol_form_vap_comp_ref": (
                    137.5,
                    pyunits.J / pyunits.mol / pyunits.K,
                ),
            },
        },
        "carbon_dioxide": {
            "type": Component,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": NIST,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (44.010e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (71.8e5, pyunits.Pa),  # [1]
                "temperature_crit": (304.1, pyunits.K),  # [1]
                "omega": 0.239,  # [1]
                "cp_mol_ig_comp_coeff": {
                    "A": (1.980e1, pyunits.J / pyunits.mol / pyunits.K),  # [1]
                    "B": (7.344e-2, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (-5.602e-5, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (1.715e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "enth_mol_form_vap_comp_ref": (-393.51, pyunits.J / pyunits.mol),  # [2]
                "entr_mol_form_vap_comp_ref": (
                    -269,
                    pyunits.J / pyunits.mol / pyunits.K,
                ),  # [2]
                "pressure_sat_comp_coeff": {
                    "A": (6.81228, None),  # [2]
                    "B": (1301.679, pyunits.K),
                    "C": (-3.494, pyunits.K),
                },
            },
        },
    },
    # Specifying phases
    "phases": {
        "Liq": {
            "type": LiquidPhase,
            "equation_of_state": Cubic,
            "equation_of_state_options": {"type": CubicType.PR},
        },
        "Vap": {
            "type": VaporPhase,
            "equation_of_state": Cubic,
            "equation_of_state_options": {"type": CubicType.PR},
        },
    },
    # Set base units of measurement
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
    # Specifying state definition
    "state_definition": FTPx,
    "state_bounds": {
        "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
        "temperature": (10, 300, 500, pyunits.K),
        "pressure": (5e-4, 1e5, 1e10, pyunits.Pa),
    },
    "pressure_ref": (101325, pyunits.Pa),
    "temperature_ref": (298.15, pyunits.K),
    # Defining phase equilibria
    "phases_in_equilibrium": [("Vap", "Liq")],
    "phase_equilibrium_state": {("Vap", "Liq"): SmoothVLE},
    "bubble_dew_method": LogBubbleDew,
    "parameter_data": {
        "PR_kappa": {
            ("bmimPF6", "bmimPF6"): 0.000,
            ("bmimPF6", "carbon_dioxide"): -0.4071,
            ("carbon_dioxide", "carbon_dioxide"): 0.000,
            ("carbon_dioxide", "bmimPF6"): 0.0206,
        }
    },
}



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Transformation to replace variables with other variables."""
# TODO: Missing docstrings
# pylint: disable=missing-class-docstring

import pyomo.environ as pyo
from pyomo.core.base.transformation import TransformationFactory
from pyomo.core.plugins.transform.hierarchy import NonIsomorphicTransformation
from pyomo.core import expr as EXPR
from pyomo.contrib.fbbt.fbbt import compute_bounds_on_expr
from pyomo.repn import generate_standard_repn
import idaes.logger as idaeslog


__author__ = "John Eslick"


_log = idaeslog.getLogger(__name__)


@TransformationFactory.register(
    "simple_equality_eliminator",
    doc="Eliminate simple equalities in the form a*x + b*y = c or a*x = b",
)
class SimpleEqualityEliminator(NonIsomorphicTransformation):
    def _get_subs(self, instance):
        subs = {}  # Substitute one var for another from a * x + b * y + c = 0
        subs_map = {}  # id -> var
        fixes = []  # fix a variable from a * x + c = 0
        cnstr = set()  # constraints to deactivate
        rset = set()  # variables used in a sub or fixed
        for c in instance.component_data_objects(pyo.Constraint, active=True):
            if (
                pyo.value(c.lower) is not None
                and pyo.value(c.lower) == pyo.value(c.upper)
                and c.body.polynomial_degree() == 1
            ):
                repn = generate_standard_repn(c.body)
                if len(repn.nonlinear_vars) != 0 or len(repn.quadratic_vars) != 0:
                    continue
                if len(repn.linear_vars) > 2:
                    continue
                elif len(repn.linear_vars) < 1:
                    _log.warning("Constraint with no vars {}: {}".format(c, c.expr))
                    continue
                b0 = repn.constant - pyo.value(c.upper)
                v0 = repn.linear_vars[0]
                a0 = repn.linear_coefs[0]
                if id(v0) in rset:
                    continue
                elif len(repn.linear_vars) == 1:
                    fixes.append((v0, -b0 / a0))
                    rset.add(id(v0))
                    cnstr.add(c)
                    continue

                v1 = repn.linear_vars[1]
                a1 = repn.linear_coefs[1]
                if id(v1) in rset:
                    continue

                subs[id(v0)] = -(b0 + a1 * v1) / a0
                cnstr.add(c)
                rset.add(id(v0))
                rset.add(id(v1))

                if self.reversible:
                    subs_map[id(v0)] = v0

                _log.debug("Sub: {} = {}".format(v0, subs[id(v0)]))

                # Use the tightest set of bounds from v0 and v1
                lb, ub = compute_bounds_on_expr(-(b0 + a0 * v0) / a1)

                if lb is not None:
                    if v1.lb is None or lb > v1.lb:
                        v1.setlb(lb)
                if ub is not None:
                    if v1.ub is None or ub < v1.ub:
                        v1.setub(ub)

        return subs, cnstr, fixes, subs_map

    def _apply_to(self, instance, max_iter=5, reversible=True):
        """
        Apply the transformation.  This is called by ``apply_to`` in the
        superclass, and should not be called directly.  ``apply_to`` takes the
        same arguments.

        Args:
            instance: A block or model to apply the transformation to

        Returns:
            None
        """
        self.reversible = reversible
        if reversible:
            self._instance = instance
            self._all_subs = []
            self._subs_map = {}
            self._expr_map = {}
            self._all_fixes = []
            self._all_deactivate = []
            self._original = {}

            # The named expressions could be changed as a side effect of the
            # constraint expression replacements, so for maximum safety, just
            # store all the expressions for Expressions
            for c in instance.component_data_objects(
                pyo.Expression,
                descend_into=True,
            ):
                self._original[id(c)] = c.expr
                self._expr_map[id(c)] = c

        nr_tot = 0
        # repeat elimination until no more can be eliminated or hit max_iter
        for i in range(max_iter):  # pylint: disable=unused-variable
            subs, cnstr, fixes, subs_map = self._get_subs(instance)

            if reversible:
                self._all_fixes += fixes
                self._all_deactivate += cnstr
                self._all_subs.append(subs)
                self._subs_map.update(subs_map)

            nr = len(cnstr)
            if nr == 0:
                break
            nr_tot += nr

            for c in cnstr:  # deactivate constraints that aren't needed
                c.deactivate()
            for v in fixes:  # fix variables that can be fixed
                v[0].fix(v[1])

            # Do replacements in Expressions, Constraints, and Objectives
            # where one var is replaced with a linear expression containing
            # another
            vis = EXPR.ExpressionReplacementVisitor(
                substitute=subs,
                descend_into_named_expressions=True,
                remove_named_expressions=False,
            )
            for c in instance.component_data_objects(
                (pyo.Constraint, pyo.Objective), descend_into=True, active=True
            ):
                if id(c) not in self._original and reversible:
                    self._original[id(c)] = c.expr
                    self._expr_map[id(c)] = c
                c.set_value(expr=vis.walk_expression(c.expr))

        _log.info("Eliminated {} variables and constraints".format(nr_tot))

    @staticmethod
    def get_logger(level=None):
        """Return the logger for this, so its a little easier to get the debugging
        output, by changing the level.
        """
        if level is not None:
            _log.setLevel(level)
        return _log

    def revert(self):
        """Revert model to pretransformation state, using substitutions to
        calculate values of variables that were removed from the problem. This
        applies to the last reversible transformation performed with this object.
        """
        try:
            instance = self._instance  # pylint: disable=unused-variable
        except AttributeError:
            _log.warning("Nothing to revert.")

        for c in self._all_deactivate:
            c.activate()
        for c in self._all_fixes:
            c[0].unfix()
        for cid, v in self._original.items():
            c = self._expr_map[cid]
            c.set_value(expr=v)

        # The problem should be back, now fill in values for the variables that
        # were removed
        for subs in reversed(self._all_subs):
            for sid in subs:
                self._subs_map[sid].value = pyo.value(subs[sid])

        del self._instance
        del self._all_subs
        del self._all_fixes
        del self._all_deactivate
        del self._original



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""This provides functions to create and check Helmholtz equation of state
parameter and expression files.
"""

import logging
import json
import numpy
import pyomo.environ as pyo

from idaes.models.properties.general_helmholtz.expressions import (
    phi_residual_types,
    phi_ideal_types,
    delta_sat_types,
    surface_tension_types,
)

_log = logging.getLogger("idaes.helmholtz_parameters")


def _parse_int_key(pairs):
    """Since json can only store dictionary keys as strings this converts the string
    keys to integers or integer tuples if possible
    """
    d = {}
    for x in pairs:
        try:
            if x[0][0] == "(" and x[0][-1] == ")":
                d[tuple(map(int, map(str.strip, x[0][1:-1].split(","))))] = x[1]
            else:
                d[int(x[0])] = x[1]
        except ValueError:
            d[x[0]] = x[1]
    return d


class WriteParameters(object):
    """This class allows users to set parameters and expressions to generate parameter
    and NL expression files that define a Helmholtz equation of state and optionally
    transport properties for a substance.
    """

    # Dictionary used to track the location of variables in NL files.
    variables = {
        "delta": 0,
        "tau": 1,
        "p": 2,
        "T": 3,
    }

    # Dictionary used to track the location of expressions in NL files.
    expressions = {
        # phi is dimensionless Helmholtz free energy
        "phii": 0,  # ideal part of phi(delta, tau)
        "phii_d": 1,  # ideal part of phi partial wrt delta
        "phii_dd": 2,  # ideal part of phi partial wrt delta and delta
        "phii_t": 3,  # ideal part of phi partial wrt tau
        "phii_tt": 4,  # ideal part of phi partial tau delta and tau
        "phii_dt": 5,  # ideal part of phi partial wrt delta and tau
        "phir": 6,  # residual part of phi(delta, tau)
        "phir_d": 7,  # residual part of phi partial wrt delta
        "phir_dd": 8,  # residual part of phi partial wrt delta and delta
        "phir_t": 9,  # residual part of phi partial wrt tau
        "phir_tt": 10,  # residual part of phi partial wrt tau and tau
        "phir_dt": 11,  # residual part of phi partial wrt delta and tau
        # Initial guess on saturated liquid and vapor reduced density
        "delta_v_sat_approx": 12,  # approximate delta_sat_vapor(tau)
        "delta_l_sat_approx": 13,  # approximate delta_sat_liquid(tau)
    }

    # A dictionary of optional expressions for transport properties, the value
    # in the dictionary is a short name used in the NL file names.
    optional_expressions = {
        "thermal_conductivity": "tcx",
        "surface_tension": "st",
        "viscosity": "visc",
    }

    def __init__(
        self,
        parameters,
    ):
        """Create a parameter writer object.  Parameters can either be specified in
        dictionary or read from a json file.

        Args:
            parameters (str|dict):  Either a dictionary of parameters or a path to
                a json file.

        Returns:
            WriteParameters: Object used to write Helmholtz EoS parameter and expression files
        """
        # A list that will store the names of defined expressions
        self.has_expression = []

        # if parameters is a string read from json
        if isinstance(parameters, str):
            with open(parameters, "r") as fp:
                parameters = json.load(fp, object_pairs_hook=_parse_int_key)
        # we may need to access parameters that are provided but not stored
        # directly as attributes, so store a link to the parameter dictionary
        self.parameters = parameters
        # Store the basic parameters as attributes.
        self.comp = parameters["comp"]
        self.R = parameters["basic"]["R"]
        self.MW = parameters["basic"]["MW"]
        self.T_star = parameters["basic"]["T_star"]
        self.rho_star = parameters["basic"]["rho_star"]
        self.Tc = parameters["basic"]["Tc"]
        self.rhoc = parameters["basic"]["rhoc"]
        self.Pc = parameters["basic"]["Pc"]
        self.Tt = parameters["basic"]["Tt"]
        self.Pt = parameters["basic"]["Pt"]
        self.rhot_l = parameters["basic"]["rhot_l"]
        self.rhot_v = parameters["basic"]["rhot_v"]
        self.P_min = parameters["basic"]["P_min"]
        self.P_max = parameters["basic"]["P_max"]
        self.rho_max = parameters["basic"]["rho_max"]
        self.T_min = parameters["basic"]["T_min"]
        self.T_max = parameters["basic"]["T_max"]
        # If provided, set the reference state offset, otherwise set to (0, 0).
        try:
            self.reference_state_offset = parameters["eos"]["reference_state_offset"]
        except KeyError:
            self.reference_state_offset = [0.0, 0.0]

        # Create the main Pyomo model to write equation of state expressions
        self.model = self.make_model("delta", "tau")

        # Optional models (Thermal Conductivity, Viscosity, and Surface Tension)
        self.model_tcx = self.make_model("delta", "tau")
        self.model_visc = self.make_model("delta", "tau")

        # Since surface tension is only valid for the two phase region, surface tension
        # is a function of temperature only
        self.model_st = self.make_model("tau")

        # Check if a predefined form of the ideal part of Helmholtz free energy is used
        try:
            phi_ideal_type = parameters["eos"]["phi_ideal_type"]
            if phi_ideal_type:
                self.add(
                    phi_ideal_types[phi_ideal_type](
                        model=self.model, parameters=parameters
                    )
                )
        except KeyError:
            phi_ideal_type = 0

        # Check if a predefined form of the residual part of Helmholtz free energy is used
        try:
            phi_residual_type = parameters["eos"]["phi_residual_type"]
            if phi_residual_type:
                self.add(
                    phi_residual_types[phi_residual_type](
                        model=self.model, parameters=parameters
                    )
                )
        except KeyError:
            phi_residual_type = 0

        # Check if predefined approximate liquid and vapor saturated density curves are used
        aux_parameters = parameters.get("aux", None)
        if aux_parameters is not None:
            delta_l_sat_parameters = parameters["aux"].get("delta_l_sat_approx", None)
            delta_v_sat_parameters = parameters["aux"].get("delta_v_sat_approx", None)
            if delta_l_sat_parameters is not None:
                etype = delta_l_sat_parameters.get("type", 0)
                if etype:
                    self.add(
                        {
                            "delta_l_sat_approx": delta_sat_types[etype](
                                model=self.model,
                                name="delta_l_sat_approx",
                                parameters=parameters,
                            ),
                        }
                    )
            if delta_v_sat_parameters is not None:
                etype = delta_v_sat_parameters.get("type", 0)
                if etype:
                    self.add(
                        {
                            "delta_v_sat_approx": delta_sat_types[etype](
                                model=self.model,
                                name="delta_v_sat_approx",
                                parameters=parameters,
                            ),
                        }
                    )

        # Check if using predefined surface tension expression
        try:
            etype = parameters["transport"]["surface_tension"]["type"]
            if etype:
                self.add(
                    {
                        "surface_tension": surface_tension_types[etype](
                            model=self.model_st, parameters=parameters
                        ),
                    }
                )
        except KeyError:  # No surface tension to add
            pass

    def calculate_pressure(self, rho, T):
        """From the expressions provided, calculate pressure from density and
        temperature. This can be used for testing and calculating the critical
        pressure based on the critical temperature and density.

        Args:
            rho (float): density in kg/m3
            T (float): temperature in K

        Returns:
            float: pressure in kPa
        """
        self.model.delta = rho / self.rho_star
        self.model.tau = self.T_star / T
        return pyo.value(rho * self.R * T * (1 + self.model.delta * self.model.phir_d))

    def calculate_enthalpy(self, rho, T):
        """From the expressions provided, calculate enthalpy from density and
        temperature. This can be used for testing.

        Args:
            rho (float): density in kg/m3
            T (float): temperature in K

        Returns:
            float: enthalpy in kJ/kg
        """
        self.model.delta = rho / self.rho_star
        self.model.tau = self.T_star / T
        return pyo.value(
            self.R
            * T
            * (
                1
                + self.model.tau * (self.model.phii_t + self.model.phir_t)
                + self.model.delta * self.model.phir_d
            )
        )

    def calculate_entropy(self, rho, T):
        """From the expressions provided, calculate entropy from density and
        temperature. This can be used for testing.

        Args:
            rho (float): density in kg/m3
            T (float): temperature in K

        Returns:
            float: entropy in kJ/kg/K
        """
        self.model.delta = rho / self.rho_star
        self.model.tau = self.T_star / T
        return pyo.value(
            self.R
            * (
                self.model.tau * (self.model.phii_t + self.model.phir_t)
                - self.model.phii
                - self.model.phir
            )
        )

    def make_model(self, *args):
        """Make a Pyomo model used to define expression NL files.  The arguments are strings
        for variables to create.  The basic parameters are also added as parameters in the
        model.

        Args:
            (str): Variables to add to the model

        Returns:
            ConcreteModel: Pyomo model with variables from args
        """
        m = pyo.ConcreteModel()
        for a in args:
            setattr(m, a, pyo.Var())
        m.R = pyo.Param(initialize=self.R)
        m.MW = pyo.Param(initialize=self.MW)
        m.T_star = pyo.Param(initialize=self.T_star)
        m.rho_star = pyo.Param(initialize=self.rho_star)
        m.Tc = pyo.Param(initialize=self.Tc)
        m.rhoc = pyo.Param(initialize=self.rhoc)
        m.Pc = pyo.Param(initialize=self.Pc)
        m.Tt = pyo.Param(initialize=self.Tt)
        m.Pt = pyo.Param(initialize=self.Pt)
        m.rhot_l = pyo.Param(initialize=self.rhot_l)
        m.rhot_v = pyo.Param(initialize=self.rhot_v)
        m.P_min = pyo.Param(initialize=self.P_min)
        m.P_max = pyo.Param(initialize=self.P_max)
        m.rho_max = pyo.Param(initialize=self.rho_max)
        m.T_min = pyo.Param(initialize=self.T_min)
        m.T_max = pyo.Param(initialize=self.T_max)
        return m

    def add(self, expressions):
        """This adds expressions to the to the object.  These expressions are written
        to NL files to be used by external functions.

        Args:
            expressions (dict): Dictionary where the key is an expression name and the
                value is a Pyomo expression.

        Returns:
            None
        """
        for name, expr in expressions.items():
            if name in self.expressions:  # check if in thermo model
                m = self.model
            elif name == "thermal_conductivity":
                m = self.model_tcx
            elif name == "surface_tension":
                m = self.model_st
            elif name == "viscosity":
                m = self.model_visc
            else:
                raise RuntimeError(f"Unknown expression {name}")
            if callable(expr):
                setattr(m, name, pyo.Objective(rule=expr))
            else:
                setattr(m, name, pyo.Objective(expr=expr))
            self.has_expression.append(name)

    def approx_sat_curves(self, trange):
        """_log.infos a table to verify that the approximate saturated density curves
        are correct.  Since the approximate curves are used as an initial guess to
        the phase equilibrium problems, they are not directly testable.

        Args:
            trange (iterable): temperature points in K

        Returns:
            tuple: list of liquid densities and list of vapor densities
        """
        _log.info(
            "\n====================================================================="
        )
        _log.info(" Check approx sat delta curves")
        _log.info(
            "====================================================================="
        )
        _log.info(
            f"{'T [K]':7s}  {'T [C]':8s}  {'~rho_l [kg/m3]':14s}  {'~rho_v [kg/m3]':14s}  {'~P [kPa]':9s}"
        )
        _log.info(
            "---------------------------------------------------------------------"
        )
        rhol_list = []
        rhov_list = []
        for T in trange:
            self.model.tau = self.model.T_star / T
            delta_l = pyo.value(self.model.delta_l_sat_approx)
            delta_v = pyo.value(self.model.delta_v_sat_approx)
            rho_l = pyo.value(delta_l * self.model.rho_star)
            rho_v = pyo.value(delta_v * self.model.rho_star)
            rhol_list.append(rho_l)
            rhov_list.append(rho_v)
            self.model.delta = delta_v
            P_v = pyo.value(
                rho_v * self.R * T * (1 + self.model.delta * self.model.phir_d)
            )
            _log.info(
                f"{T:7.3f}, {T - 273.15: 8.3f}, {rho_l:14.4f}, {rho_v:14.4f}, {P_v:9.4f}"
            )
        _log.info(
            "=====================================================================\n"
        )
        return (rhol_list, rhov_list)

    def calculate_reference_offset(self, delta, tau, s0, h0):
        """Given delta and tau for a reference state and the reference entropy and enthalpy, calculate
        the reference state offset parameters. Since temperature and density are independent of reference
        state, they can be calculated at a new reference state.

        Args:
            delta (float): reduced density at new reference state
            tau (float): 1/reduced temperature at new reference state
            s0 (float): Entropy at new reference state [kJ/kg/K]
            h0 (float): Enthalpy at new reference state [kJ/kg]

        Returns:
            (tuple): Offset for given reference state
        """
        self.model.tau = tau
        self.model.delta = delta
        s1 = pyo.value(
            self.model.tau * (self.model.phii_t + self.model.phir_t)
            - self.model.phii
            - self.model.phir
        )
        n1_off = s1 - s0
        h1 = pyo.value(
            1.0
            / self.model.tau
            * (
                1
                + self.model.tau * (self.model.phii_t + self.model.phir_t)
                + self.model.delta * self.model.phir_d
            )
        )
        n2_off = h0 - h1
        return (n1_off, n2_off)

    def write_model(self, model, model_name, expressions=None):
        """Write an NL file and create an expression and variable map for the EoS model or just a
        variable map for transport property expressions where there is only one expression per file.

        Args:
            model (ConcreteModel): a Pyomo model
            model_name (str): the model name used in the NL file

        Returns:
            tuple: NL file, expression map, variable map for EOS
        """
        nl_file, smap_id = model.write(f"{self.comp}_expressions_{model_name}.nl")
        for v in model.component_data_objects(pyo.Var):
            v.unfix()
        smap = model.solutions.symbol_map[smap_id]
        var_map = [1000] * 4
        for s, c in smap.bySymbol.items():
            if s.startswith("v"):
                j = int(s[1:])
                var_map[j] = self.variables[c.name]
        if expressions is not None:
            expr_map = [0] * len(expressions)
            for s, c in smap.bySymbol.items():
                if s.startswith("o"):
                    i = expressions[c.name]
                    j = int(s[1:])
                    expr_map[i] = j
            return nl_file, expr_map, var_map
        return nl_file, None, var_map

    def write(self, dry_run=False):
        """Write the parameter and expression files, and _log.info some diagnostics."""
        if dry_run:
            # TODO<jce> this isn't important to IDAES function, so I need to come back later
            #    and figure out how to test this.
            return

        _log.info(
            "\n======================================================================="
        )
        _log.info(f" Writing expressions for {self.comp}")
        _log.info(
            "======================================================================="
        )
        pc = self.calculate_pressure(self.rhoc, self.Tc)
        _log.info(f"Calculated Pc = {pc}, Pc given {self.Pc}")
        _log.info(f"Tc = {self.Tc}, T^* = {self.T_star}")
        _log.info(f"rhoc = {self.rhoc}, rho^* = {self.rho_star}")
        self.Pc = pc
        for name in self.expressions:
            if name not in self.has_expression:
                raise RuntimeError(f"Required expression {name} not provided.")
        nl_file, expr_map, var_map = self.write_model(
            self.model, "eos", self.expressions
        )
        param_dict = {
            "nl_file": nl_file,
            "expr_map": expr_map,
            "var_map": var_map,
            "param": {
                "R": self.R,
                "MW": self.MW,
                "T_star": self.T_star,
                "rho_star": self.rho_star,
                "Tc": self.Tc,
                "rhoc": self.rhoc,
                "Pc": self.Pc,
                "Tt": self.Tt,
                "Pt": self.Pt,
                "rhot_l": self.rhot_l,
                "rhot_v": self.rhot_v,
                "P_min": self.P_min,
                "P_max": self.P_max,
                "rho_max": self.rho_max,
                "T_min": self.T_min,
                "T_max": self.T_max,
                "reference_state_offset": self.reference_state_offset,
            },
        }

        # Add optional models
        for name, short_name in self.optional_expressions.items():
            if name in self.has_expression:
                model = getattr(self, f"model_{short_name}")
                nl_file, _, var_map = self.write_model(model, short_name)
                param_dict[f"nl_file_{short_name}"] = nl_file
                param_dict[f"var_map_{short_name}"] = var_map
                param_dict[f"have_{short_name}"] = True
            else:
                _log.warning(f"Missing optional expression {name}")
                param_dict[f"have_{short_name}"] = False

            with open(f"{self.comp}_parameters.json", "w") as f:
                json.dump(param_dict, f, indent=4)

        trng = []
        linspc = numpy.linspace(self.Tt, self.T_star, 15)
        for i, t in enumerate(linspc):
            if i != 0 and i != len(linspc) - 1:
                t = round(t, 3)
            trng.append(t)
        self.approx_sat_curves(trng)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Implementation of the formulation proposed in:

Dabadghao, V., Ghouse, J., Eslick, J., Lee, A., Burgard, A., Miller, D., Biegler, L., 2022,
A complementarity-based vapor-liquid equilibrium formulation for equation-oriented simulation
and optimization. AIChE Journal, DOI: 10.1002/aic.18029
"""

from pyomo.environ import (
    Constraint,
    Expression,
    Param,
    Set,
    units as pyunits,
    Var,
    value,
)

from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.math import smooth_min
from idaes.models.properties.modular_properties.base.utility import (
    identify_VL_component_list,
)
from idaes.models.properties.modular_properties.base.utility import (
    estimate_Tbub,
    estimate_Tdew,
)
from idaes.models.properties.modular_properties.eos.ceos import (
    calculate_equilibrium_cubic_coefficients,
)
import idaes.core.util.scaling as iscale


# Small value for initializing slack variables
EPS_INIT = 1e-4


# -----------------------------------------------------------------------------
class CubicComplementarityVLE:
    """
    Improved Vapor-Liquid Equilibrium complementarity formulation for Cubic Equations of State.
    """

    @staticmethod
    def phase_equil(b, phase_pair):
        """
        Method for constructing phase equilibrium variables and constraints
        """
        # This method is called via StateBlock.build, thus does not need clean-up
        # try/except statements
        suffix = "_" + phase_pair[0] + "_" + phase_pair[1]

        # Smooth VLE assumes a liquid and a vapor phase, so validate this
        (
            l_phase,
            v_phase,
            _,
            _,
            _,
            _,
        ) = identify_VL_component_list(b, phase_pair)

        if l_phase is None or v_phase is None:
            raise ConfigurationError(
                f"{b.params.name} - Generic Property Package phase pair {phase_pair[0]}-{phase_pair[1]} "
                "was set to use CubicComplementarityVLE formulation, however this is not a vapor-liquid pair."
            )

        try:
            lobj = b.params.get_phase(l_phase)
            ctype = lobj._cubic_type  # pylint: disable=protected-access
            cname = lobj.config.equation_of_state_options["type"].name
            vobj = b.params.get_phase(v_phase)

            if (
                ctype != vobj._cubic_type  # pylint: disable=protected-access
                or lobj.config.equation_of_state_options
                != vobj.config.equation_of_state_options
            ):
                raise ConfigurationError(
                    f"{b.params.name} - CubicComplementarityVLE formulation requires that both "
                    "phases use the same type of cubic equation of state."
                )
        except AttributeError:
            raise ConfigurationError(
                f"{b.params.name} - CubicComplementarityVLE formulation only supports cubic equations of state."
            )

        # Definition of equilibrium temperature for smooth VLE
        uom = b.params.get_metadata().default_units
        t_units = uom.TEMPERATURE
        f_units = uom.AMOUNT / uom.TIME

        vl_phase_set = Set(initialize=[phase_pair[0], phase_pair[1]])
        b.add_component("_vle_set" + suffix, vl_phase_set)

        s = Var(
            vl_phase_set,
            initialize=EPS_INIT,
            bounds=(0, None),
            doc="Slack variable for equilibrium temperature",
            units=t_units,
        )
        b.add_component("s" + suffix, s)

        # Equilibrium temperature
        def rule_teq(b):
            return (
                b._teq[phase_pair]  # pylint: disable=protected-access
                - b.temperature
                - s[v_phase]
                + s[l_phase]
                == 0
            )

        b.add_component("_teq_constraint" + suffix, Constraint(rule=rule_teq))

        # Epsilon variables will be given units of flow, as this is usually what dominates
        # the complementarity equations.
        eps_t = Param(
            default=1,
            mutable=True,
            doc="Smoothing parameter for temperature complementarity",
            units=f_units,
        )
        b.add_component("eps_t" + suffix, eps_t)

        eps_z = Param(
            default=1,
            mutable=True,
            doc="Smoothing parameter for cubic root complementarities",
            units=f_units,
        )
        b.add_component("eps_z" + suffix, eps_z)

        gp = Var(
            vl_phase_set,
            initialize=EPS_INIT,
            bounds=(0, None),
            doc="Slack variable for cubic second derivative for phase p",
            units=pyunits.dimensionless,
        )
        b.add_component("gp" + suffix, gp)

        gn = Var(
            vl_phase_set,
            initialize=EPS_INIT,
            bounds=(0, None),
            doc="Slack variable for cubic second derivative for phase p",
            units=pyunits.dimensionless,
        )
        b.add_component("gn" + suffix, gn)

        def rule_temperature_slack_complementarity(b, p):
            flow_phase = b.flow_mol_phase[p]

            return smooth_min(s[p] * f_units / t_units, flow_phase, eps_t) == 0

        b.add_component(
            "temperature_slack_complementarity" + suffix,
            Constraint(
                vl_phase_set,
                rule=rule_temperature_slack_complementarity,
            ),
        )

        def rule_cubic_second_derivative(b, p):
            p1, p2 = phase_pair

            _b, _, _ = calculate_equilibrium_cubic_coefficients(
                b, cname, ctype, p1, p2, p
            )
            z = b.compress_fact_phase[p]

            return 6 * z + 2 * _b

        b.add_component(
            "cubic_second_derivative" + suffix,
            Expression(vl_phase_set, rule=rule_cubic_second_derivative),
        )

        def rule_cubic_root_complementarity(b, p):
            der = getattr(b, "cubic_second_derivative" + suffix)
            return der[p] == gp[p] - gn[p]

        b.add_component(
            "cubic_root_complementarity" + suffix,
            Constraint(vl_phase_set, rule=rule_cubic_root_complementarity),
        )

        def rule_cubic_slack_complementarity(b, p):
            flow_phase = b.flow_mol_phase[p]
            if b.params.get_phase(p).is_vapor_phase():
                return smooth_min(gn[p] * f_units, flow_phase, eps_z) == 0
            else:
                return smooth_min(gp[p] * f_units, flow_phase, eps_z) == 0

        b.add_component(
            "cubic_slack_complementarity" + suffix,
            Constraint(vl_phase_set, rule=rule_cubic_slack_complementarity),
        )

    @staticmethod
    def calculate_scaling_factors(b, phase_pair):
        """
        Method to calculate scaling factors for phase equilibrium
        """
        suffix = "_" + phase_pair[0] + "_" + phase_pair[1]
        sf_T = iscale.get_scaling_factor(b.temperature, default=1, warning=True)

        try:
            teq_cons = getattr(b, "_teq_constraint" + suffix)
            # pylint: disable-next=protected-access
            iscale.set_scaling_factor(b._teq[phase_pair], sf_T)
            iscale.constraint_scaling_transform(teq_cons, sf_T, overwrite=False)
        except AttributeError:
            pass

    @staticmethod
    def calculate_teq(blk, pp):
        """
        Method to calculate initial guess for equilibrium temperature
        """
        # ---------------------------------------------------------------------
        # If present, initialize bubble and dew point calculations, and
        # equilibrium temperature _teq
        T_units = blk.params.get_metadata().default_units.TEMPERATURE

        (
            liquid_phase,
            vapor_phase,
            raoult_comps,
            henry_comps,
            _,
            v_only_comps,
        ) = identify_VL_component_list(blk, pp)

        if v_only_comps is None:
            if blk.is_property_constructed("temperature_bubble"):
                Tbub = value(blk.temperature_bubble[pp])
            else:
                Tbub = estimate_Tbub(
                    blk, T_units, raoult_comps, henry_comps, liquid_phase
                )
            t1 = max(value(blk.temperature), Tbub)
        else:
            t1 = value(blk.temperature)

        if v_only_comps is None:
            if blk.is_property_constructed("temperature_dew"):
                Tdew = value(blk.temperature_bubble[pp])
            else:
                Tdew = estimate_Tdew(
                    blk, T_units, raoult_comps, henry_comps, liquid_phase
                )
            t2 = min(t1, Tdew)
        else:
            t2 = t1

        blk._teq[pp].set_value(t2)  # pylint: disable=protected-access

        # ---------------------------------------------------------------------
        # Initialize sV and sL slacks
        _calculate_temperature_slacks(blk, pp, liquid_phase, vapor_phase)

        # ---------------------------------------------------------------------
        # If flash, initialize g+ and g- slacks
        _calculate_ceos_derivative_slacks(blk, pp, liquid_phase, vapor_phase)

    @staticmethod
    def phase_equil_initialization(b, phase_pair):
        """
        Method to initialize phase equilibrium
        """
        suffix = "_" + phase_pair[0] + "_" + phase_pair[1]

        for c in b.component_objects(Constraint):
            # Activate equilibrium constraints
            if c.local_name in (
                "_teq_constraint" + suffix,
                "temperature_slack_complementarity" + suffix,
                "cubic_root_complementarity" + suffix,
                "cubic_slack_complementarity" + suffix,
            ):
                c.activate()


def _calculate_temperature_slacks(b, phase_pair, liquid_phase, vapor_phase):
    suffix = "_" + phase_pair[0] + "_" + phase_pair[1]

    s = getattr(b, "s" + suffix)

    # pylint: disable-next=protected-access
    if value(b._teq[phase_pair]) > value(b.temperature):
        # pylint: disable-next=protected-access
        s[vapor_phase].set_value(value(b._teq[phase_pair] - b.temperature))
        s[liquid_phase].set_value(EPS_INIT)
    # pylint: disable-next=protected-access
    elif value(b._teq[phase_pair]) < value(b.temperature):
        s[vapor_phase].set_value(EPS_INIT)
        # pylint: disable-next=protected-access
        s[liquid_phase].set_value(value(b.temperature - b._teq[phase_pair]))
    else:
        s[vapor_phase].set_value(EPS_INIT)
        s[liquid_phase].set_value(EPS_INIT)


def _calculate_ceos_derivative_slacks(b, phase_pair, liquid_phase, vapor_phase):
    suffix = "_" + phase_pair[0] + "_" + phase_pair[1]

    gp = getattr(b, "gp" + suffix)
    gn = getattr(b, "gn" + suffix)
    der = getattr(b, "cubic_second_derivative" + suffix)

    if value(der[liquid_phase]) < 0:
        gp[liquid_phase].set_value(EPS_INIT)
        gn[liquid_phase].set_value(value(-der[liquid_phase]))
    else:
        gp[liquid_phase].set_value(EPS_INIT)
        gn[liquid_phase].set_value(EPS_INIT)

    if value(der[vapor_phase]) > 0:
        gp[vapor_phase].set_value(value(der[vapor_phase]))
        gn[vapor_phase].set_value(EPS_INIT)
    else:
        gp[vapor_phase].set_value(EPS_INIT)
        gn[vapor_phase].set_value(EPS_INIT)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

__Author__ = "John Eslick"


import pytest
import pyomo.environ as pyo
from pyomo.environ import ConcreteModel, Set, Block, Var
from pyomo.network import Port
from idaes.core.util import ModelTag, ModelTagGroup
from idaes.core.util.tags import svg_tag


@pytest.fixture()
def model():
    m = pyo.ConcreteModel()
    m.w = pyo.Var([1, 2, 3], ["a", "b"], initialize=4, units=pyo.units.kg)
    m.x = pyo.Var([1, 2, 3], initialize=5, units=pyo.units.kg)
    m.y = pyo.Var(initialize=6, units=pyo.units.s)
    m.z = pyo.Var(initialize=7)
    m.e = pyo.Expression(expr=m.w[1, "a"] / m.x[1])
    m.f = pyo.Expression(expr=m.x[1] / m.y)

    @m.Expression([1, 2, 3], ["a", "b"])
    def g(b, i, j):
        return m.w[i, j] / m.x[i] * 100

    return m


@pytest.mark.unit
def test_tag_display(model):
    m = model
    tw = ModelTag(expr=m.w, format_string="{:.3f}", doc="Tag for w")
    tf = ModelTag(expr=m.f, format_string="{:.3f}", doc="Tag for f")
    tg = ModelTag(expr=m.g, format_string="{:.1f}", doc="Tag for g", display_units="%")
    tz = ModelTag(expr=m.z, format_string="{:.1f}", doc="Tag for z")

    m.w[1, "a"] = 4.0
    m.x[1] = 8.0
    m.y = 4.0
    assert tw.doc == "Tag for w"
    assert str(tw[1, "a"]) == "4.000 kg"
    assert tw[1, "a"](units=False) == "4.000"
    assert tw[1, "a"].display(units=False) == "4.000"
    assert tw.display(units=False, index=(1, "a")) == "4.000"
    assert str(tf) == "2.000 kg/s"
    tf.str_include_units = False
    assert str(tf) == "2.000"
    assert str(tg[1, "a"]) == "50.0%"
    assert str(tg[1, "a"].display(format_string="{:.2f}")) == "50.00%"

    m.z = 7
    assert str(tz) == "7.0"


@pytest.mark.unit
def test_basic_str_float_int():
    ti = ModelTag(expr=1, format_string="{:.3f}", doc="Tag for just an int")
    tf = ModelTag(expr=2.0, format_string="{:.3f}", doc="Tag for just a float")
    ts = ModelTag(expr="hi", format_string="{:.3f}", doc="Tag for just a str")
    tn = ModelTag(expr=None, format_string="{:.3f}", doc="Tag for just None")

    assert ti.value == 1
    assert tf.value == 2.0
    assert ts.value == "hi"
    assert tn.value == None

    ti.set("new")
    tf.set("new")
    ts.set("new")
    tn.set("new")

    assert ti.value == "new"
    assert tf.value == "new"
    assert ts.value == "new"
    assert tn.value == "new"


@pytest.mark.unit
def test_tag_dict_like(model):
    m = model
    tw = ModelTag(expr=m.w, format_string="{:.3f}", doc="Tag for w")
    ty = ModelTag(expr=m.y, format_string="{:.3f}", doc="Tag for y")

    assert len(tw) == 6
    assert len(ty) == 1

    for i in ty.keys():
        assert i == None
        ty[i].set(1)

    check_w_keys = [(1, "a"), (1, "b"), (2, "a"), (2, "b"), (3, "a"), (3, "b")]
    for i, k in enumerate(tw.keys()):
        assert check_w_keys[i] == k
        tw[k].set(i)

    for i, v in enumerate(tw.values()):
        assert pyo.value(v.expression) == i

    for i, (k, v) in enumerate(tw.items()):
        assert pyo.value(v.expression) == i
        assert check_w_keys[i] == k

    assert tw[1, "a"].value == 0
    assert tw[1, "b"].value == 1
    assert tw.value[1, "b"] == 1
    assert ty.value == 1


@pytest.mark.unit
def test_tag_conditional_formatting(model):
    m = model

    tx = ModelTag(
        expr=m.x,
        format_string=lambda x: "{:,.0f}" if x >= 100 else "{:.2f}",
        doc="Tag for x",
        display_units=pyo.units.g,
    )

    tx.set(1 * pyo.units.g)
    assert str(tx[1]) == "1.00 g"
    tx.set(1 * pyo.units.kg)
    assert str(tx[1]) == "1,000 g"


@pytest.mark.unit
def test_tag_errors(model):
    m = model

    tx = ModelTag(
        expr=m.x,
        format_string=lambda x: "{:,.0f}" if x >= 100 else "{:.2f}",
        display_units=pyo.units.g,
    )
    tg = ModelTag(expr=m.g, format_string="{:,.0f}", display_units=None)
    m.x.fix(0)
    assert str(tg[1, "a"]) == "ZeroDivisionError"
    m.x.fix(None)
    assert str(tg[1, "a"]) == "None"


@pytest.mark.unit
def test_tag_display_convert(model):
    m = model
    tw = ModelTag(
        expr=m.w, format_string="{:.3f}", doc="Tag for w", display_units=pyo.units.g
    )
    tx = ModelTag(
        expr=m.x, format_string="{:.3f}", doc="Tag for x", display_units=pyo.units.g
    )
    ty = ModelTag(
        expr=m.y, format_string="{:.1f}", doc="Tag for y", display_units=pyo.units.hr
    )
    tf = ModelTag(
        expr=m.f,
        format_string="{:.1f}",
        doc="Tag for f",
        display_units=pyo.units.g / pyo.units.hr,
    )

    assert str(tf) == "3000000.0 " + str(pyo.units.g / pyo.units.hr)
    m.x[1].value = 4
    assert str(tf) == "2400000.0 " + str(pyo.units.g / pyo.units.hr)
    assert str(tx[1]) == "4000.000 " + str(pyo.units.g)
    m.x[1].value = 3
    assert str(tx[1]) == "3000.000 " + str(pyo.units.g)
    assert str(tw[1, "a"]) == "4000.000 " + str(pyo.units.g)
    assert tw[1, "a"]._index == (1, "a")
    assert tw._cache_display_value[1, "a"] == pytest.approx(4000.0)
    assert tw._cache_validation_value[1, "a"] == 4
    m.w[1, "a"].value = 1
    m.w[2, "a"].value = 2
    m.w[3, "a"].value = 3
    assert str(tw[1, "a"]) == "1000.000 " + str(pyo.units.g)
    assert str(tw[2, "a"]) == "2000.000 " + str(pyo.units.g)
    assert str(tw[3, "a"]) == "3000.000 " + str(pyo.units.g)
    assert tw._cache_display_value[1, "a"] == pytest.approx(1000.0)
    assert tw._cache_validation_value[1, "a"] == 1


@pytest.mark.unit
def test_tag_input(model):
    m = model
    tw = ModelTag(expr=m.w, format_string="{:.3f}", doc="Tag for w")
    ty = ModelTag(expr=m.y, format_string="{:.2f}", doc="Tag for y")

    tw.fix(1)
    for v in m.w.values():
        assert v.value == 1
        assert v.fixed
    tw.unfix()
    tw[:, "a"].fix(2)
    for k, v in m.w.items():
        if k[1] == "a":
            assert v.value == 2
            assert v.fixed
        else:
            assert v.value == 1
            assert not v.fixed
    tw[:, "b"].fix(4)
    for k, v in m.w.items():
        if k[1] == "b":
            assert v.value == 4

    tw.var[1, "a"] = 3
    assert issubclass(tw.var[1, "a"].ctype, pyo.Var)
    assert pyo.value(tw[1, "a"].var) == 3

    ty.fix(2)
    assert m.y.fixed
    assert pyo.value(m.y) == 2
    ty.set(4)
    assert pyo.value(m.y) == 4


@pytest.mark.unit
def test_tag_input_convert(model):
    m = model
    tw = ModelTag(
        expr=m.w, format_string="{:.3f}", doc="Tag for w", display_units=pyo.units.g
    )
    ty = ModelTag(expr=m.y, format_string="{:.2f}", doc="Tag for y")

    ty.set(1 * pyo.units.hr)
    assert str(ty) == "3600.00 s"

    ty.fix(2 * pyo.units.hr)
    assert str(ty) == "7200.00 s"
    assert m.y.fixed

    tw[:, "a"].set(3000, in_display_units=True)
    tw[:, "b"].set(2 * pyo.units.kg, in_display_units=True)
    assert pyo.value(m.w[1, "a"]) == pytest.approx(3.0)
    assert pyo.value(m.w[2, "a"]) == pytest.approx(3.0)
    assert pyo.value(m.w[3, "a"]) == pytest.approx(3.0)
    assert pyo.value(m.w[1, "b"]) == pytest.approx(2.0)
    assert pyo.value(m.w[2, "b"]) == pytest.approx(2.0)
    assert pyo.value(m.w[3, "b"]) == pytest.approx(2.0)

    tw.set_in_display_units = True
    tw[:, "a"].set(1000)
    tw[:, "b"].set(4 * pyo.units.kg)
    assert pyo.value(m.w[1, "a"]) == pytest.approx(1.0)
    assert pyo.value(m.w[2, "a"]) == pytest.approx(1.0)
    assert pyo.value(m.w[3, "a"]) == pytest.approx(1.0)
    assert pyo.value(m.w[1, "b"]) == pytest.approx(4.0)
    assert pyo.value(m.w[2, "b"]) == pytest.approx(4.0)
    assert pyo.value(m.w[3, "b"]) == pytest.approx(4.0)

    tw.set_in_display_units = False
    tw[:, "a"].set(2)
    assert pyo.value(m.w[1, "a"]) == pytest.approx(2.0)
    assert pyo.value(m.w[2, "a"]) == pytest.approx(2.0)
    assert pyo.value(m.w[3, "a"]) == pytest.approx(2.0)


@pytest.mark.unit
def test_tag_ref(model):
    m = model
    m.rw = pyo.Reference(m.w[:, "a"])
    m.ry = pyo.Reference(m.y)

    rw = ModelTag(expr=m.rw, format_string="{:.3f}", doc="Tag for rw")
    ry = ModelTag(expr=m.ry, format_string="{:.2f}", doc="Tag for ry")

    m.w[1, "a"] = 3
    assert str(rw[1]) == "3.000 kg"
    assert rw[1].is_var

    ry[:].set(2)
    assert str(ry[None]) == "2.00 s"


@pytest.mark.unit
def test_tag_data_object(model):
    m = model

    rw = ModelTag(expr=m.w[1, "a"], format_string="{:.3f}", doc="Tag for rw")

    m.w[1, "a"] = 3
    assert str(rw) == "3.000 kg"
    assert rw.is_var

    rw.set(5)
    assert str(rw) == "5.000 kg"

    rw = ModelTag(expr=m.w[1, "a"], format_string="{:.3f}", display_units=pyo.units.g)

    assert str(rw) == "5000.000 g"
    rw.set_in_display_units = True
    rw.set(5)
    assert str(rw) == "5.000 g"


@pytest.mark.unit
def test_tag_group(model):
    m = model
    g = ModelTagGroup()
    m.rw = pyo.Reference(m.w[:, "a"])
    g.add(
        "w",
        expr=m.rw,
        format_string="{:.3f}",
        doc="make sure this works",
        display_units=pyo.units.g,
    )
    g.add("x", expr=m.x, format_string="{:.3f}")
    g.add("y", expr=m.y, format_string="{:.3f}")
    g.add("z", expr=m.z, format_string="{:.3f}")
    g.add("e", expr=m.e, format_string="{:.3f}")
    g.add("f", ModelTag(expr=m.f, format_string="{:.3f}"))
    g.add("g", expr=m.g, format_string="{:.1f}", display_units="%")

    assert g["w"].doc == "make sure this works"
    g.str_include_units = True

    assert g.str_include_units
    assert g["w"].str_include_units
    assert not g.set_in_display_units
    assert not g["w"].set_in_display_units
    g["w"].fix(2)
    g["x"].fix(1)
    g["y"].fix(3)
    assert str(g["w"][1]) == "2000.000 g"
    assert str(g["x"][1]) == "1.000 kg"
    assert str(g["y"]) == "3.000 s"

    g.str_include_units = False
    assert str(g["w"][1]) == "2000.000"
    assert str(g["x"][1]) == "1.000"
    assert str(g["y"]) == "3.000"

    g.set_in_display_units = True
    g["w"][:].set(1000)
    assert str(g["w"][1]) == "1000.000"
    assert str(g["w"][2]) == "1000.000"
    assert str(g["w"][2]) == "1000.000"

    g.set_in_display_units = True
    g["w"][:].set(2 * pyo.units.kg)
    assert str(g["w"][1]) == "2000.000"
    assert str(g["w"][2]) == "2000.000"
    assert str(g["w"][2]) == "2000.000"

    g["w"][:].fix(3 * pyo.units.kg)
    assert str(g["w"][1]) == "3000.000"
    assert str(g["w"][2]) == "3000.000"
    assert str(g["w"][2]) == "3000.000"

    g["w"][:].fix(4000)
    assert str(g["w"][1]) == "4000.000"
    assert str(g["w"][2]) == "4000.000"
    assert str(g["w"][2]) == "4000.000"


@pytest.mark.unit
def test_tabulate_runs(model):
    m = model
    g = ModelTagGroup()
    g.add("w", expr=m.w, format_string="{:.3f}", display_units=pyo.units.g)
    g.add("x", expr=m.x, format_string="{:.3f}")
    g.add("y", expr=m.y, format_string="{:.3f}")
    g.add("z", expr=m.z, format_string="{:.3f}")
    g.add("e", expr=m.e, format_string="{:.3f}")
    g.add("f", ModelTag(expr=m.f, format_string="{:.3f}"))
    g.add("g", expr=m.g, format_string="{:.1f}", display_units="%")

    columns = (
        ["w", (1, "a")],
        ["w", (2, "a")],
        ["y", None],
    )

    head = g.table_heading(tags=columns, units=True)

    assert head[0] == "w[(1, 'a')] (g)"
    assert head[1] == "w[(2, 'a')] (g)"
    assert head[2] == "y (s)"

    g["w"][1, "a"].set(1 * pyo.units.g)
    g["w"][2, "a"].set(2 * pyo.units.g)
    g["y"].set(1 * pyo.units.s)

    row = g.table_row(tags=columns, units=False)
    assert row[0] == "1.000"
    assert row[1] == "2.000"
    assert row[2] == "1.000"

    row = g.table_row(tags=columns, units=True)
    assert row[0] == "1.000 g"
    assert row[1] == "2.000 g"
    assert row[2] == "1.000 s"

    row = g.table_row(tags=columns, numeric=True)
    assert row[0] == pytest.approx(1.000)
    assert row[1] == pytest.approx(2.000)
    assert row[2] == pytest.approx(1.000)

    g["w"][1, "a"].set(1 * pyo.units.g)
    g["w"][1, "b"].set(2 * pyo.units.g)
    g["w"][2, "a"].set(3 * pyo.units.g)
    g["w"][2, "b"].set(4 * pyo.units.g)
    g["w"][3, "a"].set(5 * pyo.units.g)
    g["w"][3, "b"].set(6 * pyo.units.g)

    head = g.table_heading(tags=("w", "y"), units=True)
    assert head[0] == "w[(1, 'a')] (g)"
    assert head[1] == "w[(1, 'b')] (g)"
    assert head[2] == "w[(2, 'a')] (g)"
    assert head[3] == "w[(2, 'b')] (g)"
    assert head[4] == "w[(3, 'a')] (g)"
    assert head[5] == "w[(3, 'b')] (g)"
    assert head[6] == "y (s)"

    row = g.table_row(tags=("w", "y"), units=True)
    assert row[0] == "1.000 g"
    assert row[1] == "2.000 g"
    assert row[2] == "3.000 g"
    assert row[3] == "4.000 g"
    assert row[4] == "5.000 g"
    assert row[5] == "6.000 g"
    assert row[6] == "1.000 s"


@pytest.mark.unit
def test_doc_example_and_bound(model):
    m = model
    g = ModelTagGroup()

    g["w"] = ModelTag(expr=m.w, format_string="{:.3f}")
    g["x"] = ModelTag(expr=m.x, format_string="{:.3f}", display_units=pyo.units.g)
    g["y"] = ModelTag(expr=m.y, format_string="{:.3f}")
    g["z"] = ModelTag(expr=m.z, format_string="{:.3f}")
    g["e"] = ModelTag(expr=m.e, format_string="{:.3f}")
    g["f"] = ModelTag(expr=m.f, format_string="{:.3f}")
    g["g"] = ModelTag(expr=m.g, format_string="{:.3f}")

    g.set_in_display_units = True
    g.str_include_units = False

    g["x"].set(2)
    g["x"].setlb(1)
    g["x"].setub(3)

    assert str(g["x"][1]) == "2.000"
    assert abs(g["x"][1].expression.lb - 0.001) < 1e-5  # x is in kg
    assert abs(g["x"][1].expression.ub - 0.003) < 1e-5  # x is in kg


@pytest.mark.unit
def test_tag_slice(model):
    m = model

    tw = ModelTag(
        expr=m.w[1, :],
        format_string=lambda x: "{:,.0f}" if x >= 100 else "{:.2f}",
        doc="Tag for x",
        display_units=pyo.units.g,
    )

    tw.set(1 * pyo.units.g)
    assert str(tw["a"]) == "1.00 g"
    tw.set(1 * pyo.units.kg)
    assert str(tw["b"]) == "1,000 g"



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import numpy as np

from ..atom import Atom


def isLineAtomRecord(line):
    return line[0:4] == "ATOM"


def readPointsFromPDB(filename):
    Points = []
    with open(filename, "r") as infile:
        for line in infile:
            if isLineAtomRecord(line):
                x = float(line[30:38])
                y = float(line[38:46])
                z = float(line[46:54])
                Points.append(np.array([x, y, z], dtype=float))
    return Points


def readAtomsFromPDB(filename):
    Atoms = []
    with open(filename, "r") as infile:
        for line in infile:
            if isLineAtomRecord(line):
                Atoms.append(Atom((line[12:16]).strip()))
    return Atoms


def readPointsAndAtomsFromPDB(filename):
    return readPointsFromPDB(filename), readAtomsFromPDB(filename)


def writeDesignToPDB(D, filename):
    with open(filename, "w") as outfile:
        for i in range(len(D)):
            if not (D.Contents[i] is None or D.Contents[i] == Atom()):
                outfile.write(
                    "ATOM  {:>5d} {:<4s}{:14}{:>8.3f}{:>8.3f}{:>8.3f}{:26}\n".format(
                        i,
                        D.Contents[i].Symbol,
                        "",
                        D.Canvas.Points[i][0],
                        D.Canvas.Points[i][1],
                        D.Canvas.Points[i][2],
                        "",
                    )
                )
